""" Unit tests for data-fetcher. """

from pathlib import Path

# pylint: disable=E0401, E0402, W0511
import pytest

from datafetcher.errors import (
    NGASFetchError,
    NGASServiceErrorException,
    FileErrorException,
)

# pylint: disable=C0115, C0116, C0200, C0415, R0801, R0902, R0903, R0914, R1721,
# pylint: disable=W0212, W0611, W0613, W0621, W0703, W1203

from datafetcher.datafetcher import DataFetcher
from datafetcher.utilities import (
    get_arg_parser,
    ProductLocatorLookup,
    RetrievalMode,
    Location,
    Cluster,
)


from .test_df_return_status import try_to_launch_df

# N.B. these are all in use; SonarLint just doesn't get it
from .df_pytest_utils import (
    TEST_PROFILE,
    get_mini_locations_file,
    get_locations_file,
    LOCATION_REPORTS,
    capo_settings,
    launch_datafetcher,
    settings,
    make_tempdir,
    RUN_ALL,
    confirm_retrieve_mode_copy,
    get_mini_exec_block,
)

_LOCATION_FILENAME = "locations.json"
_ASDM_XML = "ASDM.xml"
_EB_EXTERNAL_NAME = "sysstartS.58955.83384832176"

# TODO: log this in WS-179-3
print(f">>> RUNNING ALL TESTS: {RUN_ALL}")


def test_settings_setup(settings):
    """ Ensure that the test settings we're using make sense """
    assert capo_settings is not None
    assert isinstance(settings.capo_settings, dict)
    assert settings.db_settings is not None
    assert isinstance(settings.db_settings, dict)
    assert settings.test_data is not None
    assert isinstance(settings.test_data, dict)


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_nothing_retrieved_if_dry_locator(make_tempdir, settings):
    """ Simulates dry run with product locator """

    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
        "--dry-run",
    ]
    launch_datafetcher(args, settings.capo_settings)
    tempdir_files = Path(make_tempdir).iterdir()
    for file in tempdir_files:
        if not str(file).endswith(".log") and not str(file).endswith(".json"):
            pytest.fail("dry run with product locator -DID- fetch files")


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_nothing_retrieved_if_dry_file(make_tempdir, settings):
    """ dry run with locations file """

    output_dir = Path(make_tempdir)
    locations_file = get_mini_locations_file(Path(output_dir, _LOCATION_FILENAME))
    args = [
        "--location-file",
        str(locations_file),
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
        "--dry-run",
    ]
    launch_datafetcher(args, settings.capo_settings)
    tempdir_files = output_dir.iterdir()
    for file in tempdir_files:
        if not str(file).endswith(".log") and not str(file).endswith(".json"):
            pytest.fail("dry run with locations file -DID- fetch files")


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_force_flag_overwrites_existing_file(make_tempdir, capo_settings):
    top_level = Path(make_tempdir)
    location_file = get_mini_locations_file(top_level / _LOCATION_FILENAME)
    dest_dir = Path(top_level, _EB_EXTERNAL_NAME)
    dest_dir.mkdir(parents=True, exist_ok=True)

    # make a fake file to be overwritten
    fake_file = dest_dir / _ASDM_XML
    with open(fake_file, "w") as to_write:
        to_write.write("alas, my days are numbered")

    args = [
        "--location-file",
        str(location_file),
        "--profile",
        TEST_PROFILE,
        "--output-dir",
        str(top_level),
        "--force",
    ]
    launch_datafetcher(args, capo_settings)

    # go thru destination directory recursively
    # and get everybody's size
    sizes = dict()
    for file in dest_dir.rglob("*"):
        sizes[str(file)] = file.stat().st_size
    assert len(sizes) == 37
    fake_size = fake_file.stat().st_size
    assert fake_size == 9339


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_no_overwrite_without_force(make_tempdir, capo_settings):
    top_level = Path(make_tempdir)
    location_file = get_mini_locations_file(top_level / _LOCATION_FILENAME)
    locations_dict = get_mini_exec_block()
    files = locations_dict["files"]
    num_files_expected = 37
    assert len(files) == num_files_expected

    dest_dir = Path(top_level, _EB_EXTERNAL_NAME)
    dest_dir.mkdir(parents=True, exist_ok=True)

    # make a fake file to be overwritten
    fake_file = dest_dir / _ASDM_XML
    if fake_file.exists():
        fake_file.unlink()
    with open(fake_file, "w") as to_write:
        to_write.write("dang! what a kick in the rubber parts!")
    assert fake_file.stat().st_size == 38

    args = [
        "--location-file",
        str(location_file),
        "--profile",
        TEST_PROFILE,
        "--output-dir",
        str(top_level),
    ]

    with pytest.raises(FileErrorException):
        try_to_launch_df(capo_settings, args)

    retrieved = [
        file for file in top_level.rglob("*") if file.is_file() and not str(file).endswith(".json")
    ]

    # get rid of all the files we downloaded, plus the log
    [file.unlink() for file in retrieved if file.is_file() and not str(file).endswith(".json")]


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_copy_attempt_throws_fetch_error(make_tempdir, settings):
    """
    We set profile to dsoc-prod here so as to force the DF
    to try to copy rather than stream

    :param make_tempdir: temp directory fixture
    :param settings: Capo settings for datafetcher

    :return:
    """

    # N.B. can't do this import with the rest of the imports up top,
    # because test_df_return_codes not yet initialized
    from .test_df_return_status import we_are_in_docker

    if we_are_in_docker():
        # this test doesn't work in a docker container:
        # locator service URL in capo profile is None,
        # even if we write a fake props file inside the test.
        # Instead, we mock this in test_df_return_codes
        return

    prod_profile = "dsoc-prod"
    prod_props_filename = prod_profile + ".properties"
    props_file = Path(make_tempdir, prod_props_filename)

    try:
        args = [
            "--product-locator",
            settings.test_data["product_locator"],
            "--output-dir",
            str(make_tempdir),
            "--profile",
            prod_profile,
        ]
        namespace = get_arg_parser().parse_args(args)
        datafetcher = DataFetcher(namespace, settings.capo_settings)

        servers_report = datafetcher.servers_report
        confirm_retrieve_mode_copy(servers_report)

        # let's try just one file so we're not sitting here all day
        for server in servers_report:
            entry = servers_report[server]
            servers_report = {server: entry}
            datafetcher.servers_report = servers_report
            assert datafetcher.servers_report[server] is not None
            files = datafetcher.servers_report[server]["files"]
            datafetcher.servers_report[server]["files"] = [files[0]]
            break

        with pytest.raises(NGASFetchError):
            datafetcher.run()
    finally:
        if props_file.exists():
            props_file.unlink()
        assert not props_file.exists()


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_dies_with_bad_server_info(make_tempdir, settings):
    location_file = get_locations_file("VLA_BAD_SERVER")
    args = [
        "--profile",
        TEST_PROFILE,
        "--output-dir",
        str(make_tempdir),
        "--location-file",
        str(location_file),
    ]

    with pytest.raises(NGASServiceErrorException):
        launch_datafetcher(args, settings.capo_settings)


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_throws_file_err_for_existing_destination(settings):
    args = [
        "--profile",
        TEST_PROFILE,
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        "floob",
    ]
    with pytest.raises(FileErrorException):
        launch_datafetcher(args, settings.capo_settings)


@pytest.mark.skip("takes too long; re-enable with WS-179-1")
def test_gets_vlbas_from_report_file(make_tempdir, capo_settings):

    location_file = get_locations_file("VLBA_EB")
    args = [
        "--profile",
        TEST_PROFILE,
        "--output-dir",
        str(make_tempdir),
        "--location-file",
        str(location_file),
    ]
    datafetcher = DataFetcher(get_arg_parser().parse_args(args), capo_settings)
    servers_report = datafetcher.servers_report
    assert len(servers_report) == 1

    datafetcher.run()

    dest_dir = Path(make_tempdir)
    file_info_dict = dict()
    for server in servers_report.items():
        assert server[0] == "nmngas03.aoc.nrao.edu:7777"
        values = server[1]
        assert values["location"] == Location.DSOC.value
        assert values["cluster"] == Cluster.DSOC.value
        assert values["retrieve_method"] == RetrievalMode.STREAM
        file_values = values["files"]
        assert len(file_values) == 16
        for filename in file_values:
            write_fake_file(dest_dir, filename)
            file_info_dict[filename["ngas_file_id"]] = filename

    datafetcher = DataFetcher(get_arg_parser().parse_args(args), settings.capo_settings)
    datafetcher.run()

    for filename in file_info_dict:
        path = Path(dest_dir, filename)
        assert path.is_file()
        contents = path.read_text().strip()
        assert int(contents) == file_info_dict[filename]["size"]


@pytest.mark.skip("takes too long; re-enable with WS-179-1")
def test_gets_large_vla_ebs_from_report_file(make_tempdir, capo_settings):
    location_file = get_locations_file("VLA_SMALL_EB")
    args = [
        "--profile",
        TEST_PROFILE,
        "--output-dir",
        str(make_tempdir),
        "--location-file",
        str(location_file),
    ]
    datafetcher = DataFetcher(get_arg_parser().parse_args(args), capo_settings)
    servers_report = datafetcher.servers_report
    assert len(servers_report) == 2
    datafetcher.run()

    server_file_count = {
        "nmngas03.aoc.nrao.edu:7777": 0,
        "nmngas04.aoc.nrao.edu:7777": 0,
    }
    dest_dir = Path(make_tempdir)
    file_list = list()
    for server in servers_report.items():
        server_url = server[0]
        assert server_url in server_file_count.keys()
        values = server[1]
        assert values["location"] == Location.DSOC.value
        assert values["cluster"] == Cluster.DSOC.value
        assert values["retrieve_method"] == RetrievalMode.STREAM
        file_values = values["files"]
        server_file_count[server_url] += len(file_values)

        for filename in file_values:
            write_fake_file(dest_dir, filename)
            file_list.append(values)

    assert server_file_count["nmngas03.aoc.nrao.edu:7777"] == 3
    assert server_file_count["nmngas04.aoc.nrao.edu:7777"] == 41

    datafetcher = DataFetcher(get_arg_parser().parse_args(args), capo_settings)
    datafetcher.run()

    found_count = 0
    for file_info in file_list:
        for file in file_info["files"]:
            filename = file["ngas_file_id"]
            path = Path(dest_dir, filename)
            assert path.is_file()
            contents = path.read_text().strip()
            assert int(contents) == file["size"]
        found_count += 1

    assert found_count == len(file_list)


@pytest.mark.skip("takes too long; re-enable with WS-179-1")
def test_gets_images_from_report_file(make_tempdir, capo_settings):
    location_file = get_locations_file("IMG")
    args = [
        "--profile",
        TEST_PROFILE,
        "--output-dir",
        str(make_tempdir),
        "--location-file",
        str(location_file),
    ]
    datafetcher = DataFetcher(get_arg_parser().parse_args(args), capo_settings)
    servers_report = datafetcher.servers_report
    assert len(servers_report) == 2

    server_file_count = {
        "nmngas01.aoc.nrao.edu:7777": 0,
        "nmngas02.aoc.nrao.edu:7777": 0,
    }
    dest_dir = Path(make_tempdir)
    file_list = list()
    for server in servers_report.items():
        server_url = server[0]
        assert server_url in server_file_count.keys()
        values = server[1]
        assert values["location"] == Location.DSOC.value
        assert values["cluster"] == Cluster.DSOC.value
        assert values["retrieve_method"] == RetrievalMode.STREAM
        file_values = values["files"]
        server_file_count[server_url] += len(file_values)

        for filename in file_values:
            write_fake_file(dest_dir, filename)
            file_list.append(values)

    for server_url, count in server_file_count.items():
        assert count == 1

    datafetcher.run()

    found_count = 0
    for file_info in file_list:
        for file in file_info["files"]:
            filename = file["ngas_file_id"]
            path = Path(dest_dir, filename)
            assert path.is_file()
            contents = path.read_text().strip()
            assert int(contents) == file["size"]
        found_count += 1

    assert found_count == len(file_list)


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_gets_calibration_from_report_file(make_tempdir):
    location_file = get_locations_file("CALIBRATION")
    args = [
        "--profile",
        TEST_PROFILE,
        "--output-dir",
        str(make_tempdir),
        "--location-file",
        str(location_file),
    ]
    datafetcher = DataFetcher(get_arg_parser().parse_args(args), capo_settings)
    servers_report = datafetcher.servers_report
    assert len(servers_report) == 1

    fake_file = None
    file_info = None
    # (there may be more than one file; thus iteration)
    for server in servers_report.items():
        metadata = server[1]
        destination = Path(make_tempdir)
        file_info = metadata["files"][0]
        fake_file = Path(destination, file_info["ngas_file_id"])
        write_fake_file(destination, file_info)

    assert fake_file.is_file()
    contents = fake_file.read_text().strip()
    assert int(contents) == file_info["size"]


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_gets_calibration_from_locator(make_tempdir, settings):
    external_name = LOCATION_REPORTS["CALIBRATION"]["external_name"]
    product_locator = ProductLocatorLookup(settings.db_settings).look_up_locator_for_ext_name(
        external_name
    )
    args = [
        "--product-locator",
        product_locator,
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]
    namespace = get_arg_parser().parse_args(args)
    fetch = DataFetcher(namespace, settings.capo_settings)
    report_files = fetch.locations_report.files_report["files"]
    assert len(report_files) == 1

    file_spec = report_files[0]
    relative_path = file_spec["relative_path"]
    assert external_name == relative_path
    assert relative_path == file_spec["subdirectory"]
    destination = Path(make_tempdir) / relative_path
    destination.mkdir()
    write_fake_file(destination, file_spec)

    fake_file = Path(destination, file_spec["ngas_file_id"])
    assert fake_file.is_file()
    contents = fake_file.read_text().strip()
    assert int(contents) == file_spec["size"]


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_gets_gbt_data_from_locator(make_tempdir, settings):
    """ Can we cope with GBT data? """

    external_name = "AGBT17B_044_553492"
    product_locator = ProductLocatorLookup(settings.db_settings).look_up_locator_for_ext_name(
        external_name
    )
    args = [
        "--product-locator",
        product_locator,
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]
    namespace = get_arg_parser().parse_args(args)
    fetch = DataFetcher(namespace, settings.capo_settings)
    report_files = fetch.locations_report.files_report["files"]
    assert len(report_files) == 1

    file_spec = report_files[0]
    relative_path = file_spec["relative_path"]
    assert relative_path == "AGBT17B_044_01.tar"
    destination = Path(make_tempdir) / relative_path
    destination.mkdir()
    write_fake_file(destination, file_spec)

    fake_file = Path(destination, file_spec["ngas_file_id"])
    assert fake_file.is_file()
    contents = fake_file.read_text().strip()
    assert int(contents) == file_spec["size"]


def write_fake_file(destination: Path, file_info: dict):
    filename = file_info["ngas_file_id"]
    path = Path(destination, filename)
    with open(path, "w") as file:
        file.write(f'{str(file_info["size"])}\n')
