import pathlib
from typing import Dict

from .interfaces import DestinationDecorator, DeliveryContextIF
from .local import LocalDestination
from pycapo import CapoConfig


class SharedWebDestination(DestinationDecorator):
    """
    Users who do not specify a local destination receive the standard shared web destination automatically.
    This is a location in a shared filesystem where a web service is configured to allow access files.
    Otherwise, it behaves like a local destination.
    """

    def __init__(self, context: DeliveryContextIF):
        # The implementation trick here is basically that the constructor makes it look like a final destination,
        # but internally it is a decorator for a local destination with some slightly different
        # initialization parameters, which come from Capo
        capo = CapoConfig().settings("edu.nrao.archive.workflow.config.DeliverySettings")

        # eventually, we will need to plumb the username into here instead of anonymous
        username = "anonymous"

        # determine the destination directory
        self.destination_dir = pathlib.Path(capo.nraoDownloadDirectory) / username / context.token()

        # LocalDestination won't create a directory, so we'll do that here
        self.destination_dir.mkdir(parents=True)

        # we're ready to build the local destination and thread it through our super constructor
        super().__init__(LocalDestination(context, self.destination_dir))

        # generate the download URL for later usage
        self.download_url = f"{capo.nraoDownloadUrl}/{username}/{context.token()}"

    def results(self) -> Dict:
        return {"delivered_to": str(self.destination_dir), "url": self.download_url}

    def __str__(self):
        return f"shared web root at {self.destination_dir}"
