import logging
from typing import Dict, List

from messaging.router import Router, on_message

from workspaces.capability.enums import CapabilityRequestState
from workspaces.capability.helpers import Parameter
from workspaces.capability.schema import CapabilityRequest
from workspaces.capability.schema_interfaces import (
    CapabilityExecutionIF,
    CapabilityRequestIF,
)
from workspaces.capability.services.execution_manager import ExecutionManager
from workspaces.capability.services.interfaces import (
    CapabilityInfoIF,
    CapabilityServiceIF,
)
from workspaces.notification.services.interfaces import NotificationServiceIF
from workspaces.products.schema import FutureProduct
from workspaces.workflow.services.interfaces import WorkflowServiceIF

logger = logging.getLogger(__name__)


class CapabilityService(CapabilityServiceIF):
    """
    The capability service: clients access this to request capability runs
    """

    def __init__(
        self,
        capability_info: CapabilityInfoIF,
        workflow_service: WorkflowServiceIF,
        notification_service: NotificationServiceIF,
    ):
        self.message_router = Router("capability")
        self.message_router.register(self)
        self.execution_manager = ExecutionManager(
            capability_info, workflow_service, self.message_router
        )
        self.capability_info = capability_info
        self.notify = notification_service

    def create_request(
        self,
        capability_name: str,
        parameters: List[Parameter] = None,
        products: List[FutureProduct] = None,
    ) -> CapabilityRequestIF:
        """
        Create a capability request for a given capability

        :param capability_name:
        :param parameters: Parameters for request
        :param products: Products the request will create upon success
        :return: Capability request
        """
        logger.info("in create_request")
        return self.capability_info.create_capability_request(capability_name, parameters, products)

    def run_capability(self, request: CapabilityRequest) -> CapabilityExecutionIF:
        """
        Use the interface of the execution manager to run a capability
        TODO: Executions should be made from capability versions, not requests

        :param request: Capability request
        :return: Execution of the request
        """
        capability_submitted_msg = {
            "service": "capability",
            "routing_key": "capability",
            "subject": request.__json__(),
            "type": "capability-submitted",
        }
        self.message_router.send_message(**capability_submitted_msg)

        return self.execution_manager.run_capability(request)

    @on_message(type="execution-complete")
    def complete_request(self, **message: Dict):
        execution = message["subject"]
        capability_request = self.capability_info.lookup_capability_request(
            execution["capability_request_id"]
        )
        capability_request.state = CapabilityRequestState.Complete.name
        self.capability_info.save_entity(capability_request)

        capability_complete_msg = {
            "service": "capability",
            "routing_key": "capability",
            "subject": execution,
            "type": "capability-complete",
        }
        self.message_router.send_message(**capability_complete_msg)

    @on_message(service="workflow", type="delivery")
    def on_delivery(self, **message: Dict):
        """
        Catch a delivery notification and update the locations in the
        referenced execution based on it

        :param message: a delivery-type notification with a workflow request
        subject and a delivery field
        """
        logger.info("Received delivery notification: %s", message)

        # retrieve the delivery structure in the message
        delivery = message["delivery"]

        # we caught a delivery event, find the associated capability execution
        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            message["subject"]["workflow_request_id"]
        )

        # update some fields on the execution, if possible
        execution.delivery_url = delivery["url"] if "url" in delivery else None
        execution.delivery_path = delivery["delivered_to"] if "delivered_to" in delivery else None

    @on_message(type="capability-submitted")
    def notify_submitted(self, **message: Dict):
        subject = message["subject"]
        request = self.capability_info.lookup_capability_request(subject["id"])
        self.notify.notify_submitted(request)

    @on_message(type="capability-complete")
    def notify_complete(self, **message: Dict):
        subject = message["subject"]
        request = self.capability_info.lookup_capability_request(subject["id"])
        self.notify.notify_complete(request)
